use std::sync::mpsc;

use xcb;
use mpd;

use config::Config;
use ui::context::Context;
use ui::x11;

mod title;
mod tray;
mod wm;
mod spacer;
mod music;
mod time;
mod temperature;
mod disk;
mod battery;
mod netspeed;

pub fn by_type(name: &str) -> Option<WidgetConstructor> {
    match name {
        "title" => Some(title::title),
        "tray" => Some(tray::tray),
        "spacer" => Some(spacer::spacer),
        "mpd" => Some(music::mpd),
        "bspwm" => Some(wm::bspwm),
        "time" => Some(time::time),
        "temperature" => Some(temperature::temperature),
        "disk" => Some(disk::disk),
        "battery" => Some(battery::battery),
        "netspeed" => Some(netspeed::netspeed),
        _ => None
    }
}

pub type MessageSender = mpsc::Sender<Message>;

pub enum Message {
    Update,
    Quit,
    MousePress(u16),
    BspwmEvent(String),
    MpdEvent(mpd::status::State, Option<mpd::song::Song>),
    XcbEvent(xcb::GenericEvent)
}

pub enum Update {
    Nothing,
    Rerender,
    Relayout,
    Quit
}

pub trait Widget {
    fn init(&mut self) {}
    fn render(&mut self, x: u16, width: u16);
    fn width(&mut self) -> u16;
    fn fit_width(&self) -> bool {
        false
    }
    fn handle_event(&mut self, _event: &Message) -> Update {
        Update::Nothing
    }
    fn finish(&mut self) {}
}

#[derive(Clone)]
pub struct WidgetParams {
    pub id: String,
    pub context: Context,
    pub tx: MessageSender,
    pub config: Config
}

impl WidgetParams {
    fn conn(&self) -> x11::Connection {
        self.context.conn.clone()
    }

    fn window(&self) -> xcb::Window {
        self.context.window
    }
}

pub type WidgetConstructor = fn(WidgetParams) -> Box<Widget>;
