use std::fs::File;
use std::io::prelude::*;
use std::io;
use std::time::Instant;

use style;
use ui::context::Context;
use widgets::{Message, Update, Widget, WidgetParams};

struct StatFiles {
    rx: File,
    tx: File
}

struct Stats {
    rx: i64,
    tx: i64
}

pub struct NetSpeed {
    context: Context,
    files: Vec<StatFiles>,
    stats: Option<Stats>,
    status: String,
    last_time: Instant
}

pub fn netspeed(params: WidgetParams) -> Box<Widget> {
    let config: NetSpeedConfig = params.config.try_into().unwrap();
    let files: Vec<StatFiles> = config.devices.iter()
        .flat_map(|dev| open_stats(&dev).ok())
        .collect();
    let widget = NetSpeed {
        context: params.context,
        files: files,
        stats: None,
        status: "0 K↓ 0 K↑".to_string(),
        last_time: Instant::now()
    };
    Box::new(widget)
}

impl Widget for NetSpeed {
    fn render(&mut self, x: u16, width: u16) {
        style::render(&self.context, "", &self.status, x, width);
    }

    fn width(&mut self) -> u16 {
        style::width(&self.context, "", &self.status)
    }

    fn handle_event(&mut self, event: &Message) -> Update {
        match event {
            &Message::Update => {
                let now = Instant::now();

                let stats = self.files
                    .iter_mut()
                    .flat_map(|file| read_stats(file).ok())
                    .fold(Stats { rx: 0, tx: 0 }, |acc, elem| Stats {
                        rx: acc.rx + elem.rx,
                        tx: acc.tx + elem.tx
                    });

                let diff_time = now.duration_since(self.last_time);
                let diff_time = diff_time.as_secs() as f64 + diff_time.subsec_nanos() as f64 * 1e-9;

                let rate = match self.stats.as_ref() {
                    None=> Stats { rx: 0, tx: 0 },
                    Some(pstats) => {
                        let rx = (stats.rx - pstats.rx) as f64 / diff_time;
                        let tx = (stats.tx - pstats.tx) as f64 / diff_time;
                        Stats { rx: rx as i64, tx: tx as i64 }
                    }
                };

                self.last_time = now;
                self.stats = Some(stats);
                self.status = format!("{}↓ {}↑", format_bytes(rate.rx), format_bytes(rate.tx));
                Update::Relayout
            },
            _ => Update::Nothing
        }
    }
}

fn format_bytes(bytes: i64) -> String {
    let kib = bytes >> 10;
    if kib > 1024 {
        format!("{:.1} M", (kib as f32) / 1024.0)
    }
    else {
        format!("{} K", kib)
    }
}

fn open_stats(device: &str) -> Result<StatFiles, io::Error> {
    let path = format!("/sys/class/net/{}/statistics", device);
    let rx_file = try!(File::open(format!("{}/rx_bytes", path)));
    let tx_file = try!(File::open(format!("{}/tx_bytes", path)));
    Ok(StatFiles {
        rx: rx_file,
        tx: tx_file
    })
}

fn read_stats(files: &mut StatFiles) -> Result<Stats, io::Error> {
    Ok(Stats {
        rx: read_bytes(&mut files.rx),
        tx: read_bytes(&mut files.tx)
    })
}

fn read_bytes(f: &mut File) -> i64 {
    let mut s = String::new();
    assert!(f.read_to_string(&mut s).is_ok());
    let i : i64 = s.trim().parse().unwrap();
    assert!(f.seek(io::SeekFrom::Start(0)).is_ok());
    i
}

#[derive(Deserialize)]
struct NetSpeedConfig {
    devices: Vec<String>
}
